//--------------------------------------------------------------------------d
//
//     PCL-836 Driver
//
//--------------------------------------------------------------------------
#include <dos.h>
#include <stdio.h>
#include <bios.h>
#include <conio.h>
#include <stdlib.h>
#include "..\driver.h"

// Global variables
USHORT usBaseAddr[10] = {0,0,0,0,0,0,0,0,0,0};  // save base address location
USHORT usPreState[10][2];     // digital output previous state
USHORT usFirstCount[10][6];   // first read counter value for event counter
ULONG  ulOverflow[10][6];     // overflow flag for event counter
USHORT usPreCount[10][6];     // previous count state for event counter
USHORT usFoutCtrReg[10];      // frequency output control register
USHORT usModeCtrReg[10][6];   // control clock input mode control register

// Function prototypes
ULONG DoSection(USHORT ,USHORT, USHORT ,USHORT);
ULONG DiSection(USHORT, USHORT, USHORT *);
ULONG GetDoStatus(USHORT, USHORT, USHORT *);

//-------------------------------------------------------------------
// Function    : DeviceOpen
// PURPOSE     : Retrieves the parameters pertaining to the device
//               operation from Global Variable , and store it for
//               quick reference. This function must be called
//               before any other functions
// Parameters  : DeviceNumber (IN), BaseAddr(OUT)
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG DeviceOpen(USHORT DeviceNumber, USHORT BaseAddr)
{
    USHORT usIndex;
    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (BaseAddr == NULL)
        return ((ULONG) InvalidBaseAddr);

    // Is BaseAddr in use ?
    if (usBaseAddr[DeviceNumber] != NULL)
        return ((ULONG) DeviceIsBusy); // If you have this error, you
                                       // can call DeviceClose to reset it.

    // everything is OK then initiate some Global variables by DeviceNumber
    usPreState[DeviceNumber][0] = 0;   // for digital output port 0
    usPreState[DeviceNumber][1] = 0;   // for digital output port 1

    for (usIndex=0 ; usIndex < MaxCounterNumber ; usIndex++)
    {
        // use by event counter
        usFirstCount[DeviceNumber][usIndex] = 0;
        ulOverflow[DeviceNumber][usIndex]   = 0;
        usPreCount[DeviceNumber][usIndex]   = 0;
        usFoutCtrReg[DeviceNumber]          = 0;
        usModeCtrReg[DeviceNumber][usIndex] = 0;
    }

    // set Global variable by DeviceNumber
    usBaseAddr[DeviceNumber] = BaseAddr;

    // return success
    return ((ULONG) NULL);
}

//------------------------------------------------------------------
// Function    : DeviceClose
// PURPOSE     : Resets Device by DeviceNumber.
// Parameters  : DeviceNumber (IN)
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
// CounterReset
//------------------------------------------------------------------
ULONG DeviceClose(USHORT DeviceNumber)
{
    USHORT usIndex;
    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // Reset counters by DeviceNumber
    for (usIndex=0 ; usIndex < MaxCounterNumber ; usIndex++)
        CounterReset(DeviceNumber, usIndex);

    // Reset Golbal variable
    usBaseAddr[DeviceNumber] = 0;

    return ((ULONG) NULL);
}

//------------------------------------------------------------------
// Function    : CounterConfig
// PURPOSE     : This function config the counter mode of specified counter
//               (default is positive edge without filter)
// Parameters  : DeviceNumber(IN), counter(IN), CounterEdge(IN)
//               MaxInFreq(IN)
//               312K is high limitation of MaxInFreq.
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
extern ULONG CounterConfig(USHORT DeviceNumber, USHORT counter,
                           USHORT CounterEdge, FLOAT MaxInFreq)
{
    USHORT usCount;
    USHORT usCtrByte, usLowByte, usHighByte;
    ULONG  retCode;
    FLOAT  fFout;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxCounterNumber)
        return((ULONG) InvalidCounterChannel);

    // check MaxInFreq is out of range ?
    if (MaxInFreq > (FLOAT)312000.0)
        return((ULONG) InvalidFreqRange);

    usModeCtrReg[DeviceNumber][counter] = 0x00;

    if (CounterEdge == 1)   // this mode is negative edge
        usModeCtrReg[DeviceNumber][counter] = 0x04;

    if (MaxInFreq != (FLOAT)0.0)
    {
       // set control with digital filter
       usModeCtrReg[DeviceNumber][counter] |= 0x01;

       // set filter wide
       fFout = MaxInFreq*8*2;

       retCode = FreqOutStart(DeviceNumber, counter, fFout);
       if (retCode != 0L)
          return(retCode);
    }

    return ((ULONG) 0);
}

//------------------------------------------------------------------
// Function    : CounterPulseStart
// PURPOSE     : Configures the specified counter for pulse
//               output and starts the counter.
// Parameters  : DeviceNumber(IN), counter(IN),
//               period(IN)  in ms
//               UpCycle(IN) in ms
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG CounterPulseStart(USHORT DeviceNumber,USHORT counter,
    FLOAT period, FLOAT UpCycle)
{
    USHORT usBase;
    DWORD  divisor;
    USHORT usCtrByte, usLowByte, usHighByte;
    FLOAT  fUserHz;
    FLOAT fFout;
    USHORT usCount0, usCount3;
    ULONG retCode;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxPWMNumber)
        return((ULONG) InvalidCounterChannel);

    // check Period is out of range ?
    if ((period  > (FLOAT)10000)   || (UpCycle > (FLOAT)10000) ||
        (period == 0)              || (UpCycle == 0)            ||
        (UpCycle >= period))
        return((ULONG) InvalidInputParam);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    /* configure counter for external clocking */
    outportb(usBase+counter+18, 0x03);

    // configures counter 0 and counter 3 (for PWM channel 0)
    // configures counter 1 and counter 4 (for PWM channel 1)
    // configures counter 2 and counter 5 (for PWM channel 2)

    fUserHz = (FLOAT)((FLOAT)1.0/period)*1000;

    if (fUserHz > (FLOAT)2500000.0)  // check out of 2.5Mhz
        return((ULONG) InvalidFreqRange);

    // FOUT select
    if (fUserHz >= (FLOAT)F100HZ)
        fFout = (FLOAT)5000000.0;
    else if (fUserHz >= (FLOAT)F10HZ)
        fFout = (FLOAT)500000.0;
    else if (fUserHz >= (FLOAT)F1HZ)
        fFout = (FLOAT)50000.0;
    else if (fUserHz >= (FLOAT)F0_1HZ)
        fFout = (FLOAT)5000.0;

    usCount3 = (USHORT)(fFout/fUserHz);

   // FOUT enable
   retCode = FreqOutStart(DeviceNumber, counter, fFout);
   if (retCode != 0L)
        return(retCode);

   // set counter 3 (or 4, 5)
   usCtrByte = (counter%3) << 6;
   usCtrByte |= 0x34;
   usLowByte = (USHORT)(usCount3 & 0x000000FF);
   usHighByte = (USHORT)((usCount3 &  0x0000FF00) >> 8);

   outportb(usBase+counter+21, 0x02);        /* set internal clock */
   outportb(usBase+15, usCtrByte);
   outportb(usBase+counter+12, usLowByte);       /* write low byte */
   outportb(usBase+counter+12, usHighByte);      /* write high byte */

   // set counter 0 (or 1, 2)
   usCount0 = usCount3*((FLOAT)(period-UpCycle)/period);

   usCtrByte = (counter%3) << 6;
   usCtrByte |= 0x3a;
   usLowByte = (USHORT)(usCount0 & 0x000000FF);
   usHighByte = (USHORT)((usCount0 &  0x0000FF00) >> 8);

   outportb(usBase+counter+18, 0x03);        /* set internal clock */
   outportb(usBase+11, usCtrByte);
   outportb(usBase+counter+8, usLowByte);       /* write low byte */
   outportb(usBase+counter+8, usHighByte);      /* write high byte */


    return 0L;
}

//------------------------------------------------------------------
// Function    : FreqOutStart
// PURPOSE     : Configures the specified counter for frequency output
//               and starts the counter. (internal base clock is 10MHZ)
// Parameters  : DeviceNumber(IN), counter(IN), fout(IN)
//               fout range is from 153HZ to 5MHZ
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG FreqOutStart(USHORT DeviceNumber, USHORT counter, FLOAT fout)
{
    USHORT usBase, usCount;
    USHORT usCtrByte, usLowByte, usHighByte;
    FLOAT  fBalance;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxCounterNumber)
        return((ULONG) InvalidCounterChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    // check input fout is out of range ?
    if (( fout > (float)5000000.0) ||  ( fout < (float)153.0))
        return((ULONG) InvalidFreqRange);

    fBalance = (float)10000000.0f / fout;
    usCount = (USHORT) fBalance;

    if (((FLOAT)fBalance*10 - (FLOAT)usCount*10) > 4)
        usCount ++;

    usLowByte = (USHORT)(usCount & 0x000000FF);
    usHighByte = (USHORT)((usCount &  0x0000FF00) >> 8);

    usCtrByte = (counter%3) << 6;
    usCtrByte |= 0x36;

    if (counter < 3)
    {
       outportb(usBase+3, usCtrByte);     // Mode 3
       outportb(usBase+counter, usLowByte);       /* write low byte */
       outportb(usBase+counter, usHighByte);      /* write high byte */
    }
    else
    {
       outportb(usBase+7, usCtrByte);     // Mode 3
       outportb(usBase+(counter%3)+4, usLowByte);  /* write low byte */
       outportb(usBase+(counter%3)+4, usHighByte); /* write high byte */
    }

    // enable frequency output control reg.
    usCtrByte = 0x01;
    usCtrByte = usCtrByte << counter;
    usFoutCtrReg[DeviceNumber] |= usCtrByte;

    outportb(usBase+24, usFoutCtrReg[DeviceNumber]);

    return 0L;
}

//------------------------------------------------------------------
// Function    : FreqOutStop
// PURPOSE     : This function stop the Frequency out of specified counter
// Parameters  : DeviceNumber(IN), counter(IN)
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG FreqOutReset(USHORT DeviceNumber, USHORT counter)
{
    USHORT usBase, usCtrByte;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxCounterNumber)
        return((ULONG) InvalidCounterChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    usCtrByte = (counter%3) << 6;
    usCtrByte |= 0x39;

    // stop counting by setting counter chan for mode 5
    if (counter < 3)  // configure 8254 chip 1
    {
       outportb(usBase+3, usCtrByte);
       outportb(usBase+counter, 0x0);       /* write low byte */
       outportb(usBase+counter, 0x0);       /* write high byte */
    }
    else              // configure 8254 chip 2
    {
       outportb(usBase+7, usCtrByte);
       outportb(usBase+(counter%3)+4, 0x0);  /* write low byte */
       outportb(usBase+(counter%3)+4, 0x0);  /* write high byte */
    }

    // disable frequency output control reg.
    usCtrByte = 0x01;
    usCtrByte = usCtrByte << counter;

    usFoutCtrReg[DeviceNumber] &= (~usCtrByte);

    outportb(usBase+24, usFoutCtrReg[DeviceNumber]);

    return 0L;
}

//------------------------------------------------------------------
// Function    : CounterEventStart
// PURPOSE     : Configures the specified counter for an event-counting
//               operation and starts the counter.
//               The function uses 8254 mode 0 and initialize counter is
//               0xffff
// Parameters  : DeviceNumber(IN), counter(IN)
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG CounterEventStart(USHORT DeviceNumber, USHORT counter)
{
    USHORT usBase;
    USHORT usCtrByte, usLowByte, usHighByte, usCurrentCount;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxCounterNumber)
        return((ULONG) InvalidCounterChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    // initial Global variables for event counter
    ulOverflow[DeviceNumber][counter] = 0;
    usFirstCount[DeviceNumber][counter] = 1;

    // configure counter for external clocking
    outportb(usBase+counter+18, usModeCtrReg[DeviceNumber][counter]);

    // set control byte for  control register & set initial value
    usCtrByte = (counter%3) << 6;
    usCtrByte |= 0x30;

    if (counter < 3)  // configure 8254 chip 3
    {
       outportb(usBase+11, usCtrByte);
       outportb(usBase+counter+8, 0xff);       /* write low byte */
       outportb(usBase+counter+8, 0xff);       /* write high byte */
    }
    else              // configure 8254 chip 4
    {
       outportb(usBase+15, usCtrByte);
       outportb(usBase+(counter%3)+12, 0xff);  /* write low byte */
       outportb(usBase+(counter%3)+12, 0xff);  /* write high byte */
    }

    // Read first counter
    usCtrByte = (counter%3) << 6;
    usCtrByte |= 0;      /* set control byte for read mode */

    if (counter < 3)  // configure 8254 chip 3
    {
       outportb(usBase+11, usCtrByte);
       usLowByte  = inportb(usBase+counter+8);      /* read low byte */
       usHighByte = inportb(usBase+counter+8);      /* read high byte */
    }
    else              // configure 8254 chip 4
    {
       outportb(usBase+15, usCtrByte);
       usLowByte  = inportb(usBase+(counter%3)+12); /* read low byte */
       usHighByte = inportb(usBase+(counter%3)+12); /* read high byte */
    }

    usHighByte <<= 8;
    usCurrentCount = usHighByte | usLowByte;

    usPreCount[DeviceNumber][counter] = usCurrentCount;

    return ((ULONG) NULL);
}

//------------------------------------------------------------------
// Function    : CounterEventRead
// PURPOSE     : Reads the current counter total without disturbing
//               the counting process and returns the count and
//               overflow conditions. 
//               When your counter overs 32-BIT , its overflow will be
//               set to 1.
// Parameters  : DeviceNumber(IN), counter(IN), overflow(OUT), count(OUT)
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG CounterEventRead(USHORT DeviceNumber,USHORT counter,
    USHORT *overflow, ULONG *count)
{
    USHORT usBase;
    USHORT usCtrByte, usLowByte, usHighByte, usCurrentCount;
    ULONG  c_data, carry;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxCounterNumber)
        return((ULONG) InvalidCounterChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    // Read value
    usCtrByte = (counter%3) << 6;
    usCtrByte |= 0;      /* set control byte for read mode */

    if (counter < 3)  // configure 8254 chip 3
    {
       outportb(usBase+11, usCtrByte);
       usLowByte  = inportb(usBase+counter+8);      /* read low byte */
       usHighByte = inportb(usBase+counter+8);      /* read high byte */
    }
    else              // configure 8254 chip 4
    {
       outportb(usBase+15, usCtrByte);
       usLowByte  = inportb(usBase+(counter%3)+12); /* read low byte */
       usHighByte = inportb(usBase+(counter%3)+12); /* read high byte */
    }

    usHighByte <<= 8;
    usCurrentCount = usHighByte | usLowByte;

	// still wait for first valid count
    if(usFirstCount[DeviceNumber][counter])
	{
        if (usPreCount[DeviceNumber][counter] == usCurrentCount)
		{
			*count = 0;
            *overflow = 0;
			return 0L;
		}
		else
		{
            usFirstCount[DeviceNumber][counter] = 0;  // set not first counter
            usPreCount[DeviceNumber][counter] = 65535;  // initial previous count
		}
	}

    c_data = ((ULONG)65536 - (ULONG)usCurrentCount); /* convert to up count */

    /* check if 16-bit counter register of 8253 is over flow ?    */
    if(usPreCount[DeviceNumber][counter] < usCurrentCount)  // over flow
        ulOverflow[DeviceNumber][counter] += 1;   /* increament overflow by 1 */

    carry = ((ULONG)ulOverflow[DeviceNumber][counter]) << 16; /* shift overflow 16 bits */
    c_data |= carry;        /* convert 16-bit to 32-bit */

    // check 32bit overflow ?
    *overflow = 0;  		// clear overflow flag
    if (ulOverflow[DeviceNumber][counter] > 63335L)
        *overflow = 1;

    *count = c_data;    // return data

    usPreCount[DeviceNumber][counter] = usCurrentCount;   // save data

    return 0L;
}

//------------------------------------------------------------------
// Function    : CounterFreqStart
// PURPOSE     : Configures the specified counter for frequency
//               measurement and starts the counter.
// Parameters  : DeviceNumber(IN), counter(IN)
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG CounterFreqStart(USHORT DeviceNumber,USHORT counter)
{
    USHORT usBase;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxCounterNumber)
        return((ULONG) InvalidCounterChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    /* configure counter for external clocking */
    outportb(usBase+counter+18, usModeCtrReg[DeviceNumber][counter]);

    return 0L;
}

//------------------------------------------------------------------
// Function    : CounterFreqRead
// PURPOSE     : Reads the frequency measurement.
// Parameters  : DeviceNumber(IN), counter(IN), FreqLevel(IN), freq(OUT)
//               FreqLevel
//                 = MAX1K_65K(0),     RANGE=   1 -- 65KHZ (need    1s)
//                 = MAX10K_650K(1),   RANGE= 10K -- 650KHZ(need 100ms)
//                 = MAX100K_6500K(2), RANGE=100K -- 6.5MHZ(need  10ms)
//
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG CounterFreqRead(USHORT DeviceNumber,USHORT counter,
    USHORT FreqLevel, FLOAT * freq)
{
    USHORT usBase;
    USHORT usDelay;
    USHORT usCtrByte, usCtr1Byte;
    USHORT usLow1Byte, usHigh1Byte, usCurrentCount1;
    USHORT usLow2Byte, usHigh2Byte, usCurrentCount2;
    FLOAT  fSampleRate;

    usLow1Byte = 0;
    usHigh1Byte = 0;
    usLow2Byte = 0;
    usHigh2Byte = 0;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxCounterNumber)
        return((ULONG) InvalidCounterChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    // Prepare some thing for BIOS function call
    switch(FreqLevel)
    {
        case MAX1_65K:
            usDelay = 1000;
            fSampleRate = 1;
            break;
        case MAX10K_650K:
            usDelay = 100;
            fSampleRate = 10;
            break;
        case MAX100K_6500K:
            usDelay = 10;
            fSampleRate = 100;
            break;
        default:
            usDelay = 1000;
            fSampleRate = 1;
            break;
    }

    // first :  set control byte for  control register
    usCtrByte = (counter%3) << 6;
    usCtr1Byte = usCtrByte | 0x30;

    if (counter < 3)  // configure 8254 chip 3
    {
       outportb(usBase+11, usCtr1Byte);
       outportb(usBase+counter+8, 0xff);       /* write low byte */
       outportb(usBase+counter+8, 0xff);       /* write high byte */
    }
    else              // configure 8254 chip 4
    {
       outportb(usBase+15, usCtr1Byte);
       outportb(usBase+(counter%3)+12, 0xff);  /* write low byte */
       outportb(usBase+(counter%3)+12, 0xff);  /* write high byte */
    }

    // delay 1/2 time
    delay(usDelay/2);

    // second: read counter
    usCtr1Byte = usCtrByte | 0;      /* set control byte for read mode */

    if (counter < 3)  // configure 8254 chip 3
    {
       outportb(usBase+11, usCtr1Byte);
       usLow1Byte  = inportb(usBase+counter+8);      /* read low byte */
       usHigh1Byte = inportb(usBase+counter+8);      /* read high byte */
    }
    else              // configure 8254 chip 4
    {
       outportb(usBase+15, usCtr1Byte);
       usLow1Byte  = inportb(usBase+(counter%3)+12); /* read low byte */
       usHigh1Byte = inportb(usBase+(counter%3)+12); /* read high byte */
    }

    // second: delay time 1/2
    delay(usDelay/2);

    // third: read counter again
    usCtr1Byte = usCtrByte | 0;      /* set control byte for read mode */

    if (counter < 3)  // configure 8254 chip 3
    {
       outportb(usBase+11, usCtr1Byte);
       usLow2Byte  = inportb(usBase+counter+8);      /* read low byte */
       usHigh2Byte = inportb(usBase+counter+8);      /* read high byte */
    }
    else              // configure 8254 chip 4
    {
       outportb(usBase+15, usCtr1Byte);
       usLow2Byte  = inportb(usBase+(counter%3)+12); /* read low byte */
       usHigh2Byte = inportb(usBase+(counter%3)+12); /* read high byte */
    }

    // forth : ? frequency
    usHigh1Byte <<= 8;
    usCurrentCount1 = usHigh1Byte | usLow1Byte;
    usHigh2Byte <<= 8;
    usCurrentCount2 = usHigh2Byte | usLow2Byte;

   if (usCurrentCount1 == usCurrentCount2)
   {
      *freq = 0;
      return 0L;
   }

  if (usCurrentCount2 > usCurrentCount1)
      return((ULONG) FreqMeasurementFailed);

    *freq = (FLOAT)(((FLOAT)65536 - (FLOAT)usCurrentCount2)
            * fSampleRate);

    return 0L;
}

//------------------------------------------------------------------
// Function    : CounterReset
// PURPOSE     : Turns off the spefified counter operation. This function
//               supports the boards with the timer/counter chip.
// Parameters  : DeviceNumber(IN), counter(IN)
// Return      : NULL (success)
//				 ErrorCode (failure)
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG CounterReset(USHORT DeviceNumber,USHORT counter)
{
    USHORT usBase, usCtrByte;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (counter > MaxCounterNumber)
        return((ULONG) InvalidCounterChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    usCtrByte = (counter%3) << 6;
    usCtrByte = usCtrByte | 0x39;

    // stop counting by setting counter chan for mode 5
    if (counter < 3)  // configure 8254 chip 3
    {
       outportb(usBase+11, usCtrByte);
       outportb(usBase+counter+8, 0x0);       /* write low byte */
       outportb(usBase+counter+8, 0x0);       /* write high byte */
    }
    else              // configure 8254 chip 4
    {
       outportb(usBase+15, usCtrByte);
       outportb(usBase+(counter%3)+12, 0x0);  /* write low byte */
       outportb(usBase+(counter%3)+12, 0x0);  /* write high byte */
    }

    return 0L;
}

//------------------------------------------------------------------
// Function    : DioReadPortByte()
// PURPOSE     : Returns digital input data from the specified digital
//               I/O port.
// Parameters  : DeviceNumber(IN)
//               port(IN)
//				 value(OUT)
// Return      : TRUE (success), FALSE (failure)
//
// Call/Called procedure cross reference :
// Call        Called		     Explanation
// DiSection
//------------------------------------------------------------------
ULONG DioReadPortByte(USHORT DeviceNumber, USHORT port,USHORT * value)
{
    return(DiSection(DeviceNumber, port, (USHORT *)value));
}

//------------------------------------------------------------------
// Function    : DioWritePortByte()
// PURPOSE     : Writes digital output data to the specified the port.
//               I/O port.
// Parameters  : DeviceNumber(IN)
//               port(IN)
//				 mask(IN)
//				 state(IN)
// Return      : TRUE (success), FALSE (failure)
//
// Call/Called procedure cross reference :
// Call        Called		     Explanation
// DoSection
//------------------------------------------------------------------
ULONG DioWritePortByte(USHORT DeviceNumber,
    USHORT port, USHORT mask, USHORT state)
{
    return(DoSection(DeviceNumber, port, mask, state));
}

//------------------------------------------------------------------
// Function    : DioReadBit()
// PURPOSE     : Returns the bit state of digital input from the specified
//               digital I/O port.
// Parameters  : DeviceNumber(IN)
//               port(IN)
//				 bit(IN)
//				 state(OUT)
// Return      : TRUE (success), FALSE (failure)
//
// Call/Called procedure cross reference :
// Call        Called		     Explanation
// DiSection
//------------------------------------------------------------------
ULONG DioReadBit(USHORT DeviceNumber,
      USHORT port, USHORT bit, USHORT  * state)
{
    USHORT usMask;
    ULONG  retCode;

    if (bit > 7)
        return((ULONG) InvalidInputParam);

    retCode = DiSection(DeviceNumber, port, (USHORT *)state);

    if (retCode == 0L)
    {
        usMask = 0x01;
        usMask = usMask << bit;
        *state = *state & usMask;

        if (*state != 0)
            *state = 1;
    }

    return((ULONG) retCode);
}

//------------------------------------------------------------------
// Function    : DioWriteBit()
// PURPOSE     : Writes digital output data to the specified bit.
//               I/O port.
// Parameters  : DeviceNumber(IN)
//               port(IN)
//				 state(IN)
// Return      : TRUE (success), FALSE (failure)
//
// Call/Called procedure cross reference :
// Call        Called		     Explanation
// DoSection
//------------------------------------------------------------------
ULONG DioWriteBit(USHORT DeviceNumber,
    USHORT port, USHORT bit, USHORT state)
{
    USHORT usMask;
    ULONG retCode;

    if (bit > 7)
        return((ULONG) InvalidInputParam);

    usMask = 0x01;
    usMask = usMask << bit;
    state = state << bit;

    return((ULONG) DoSection(DeviceNumber, port, usMask, state));
}
//------------------------------------------------------------------
// Function    : DioGetCurrentDOByte()
// PURPOSE     : Returns digital input data from the specified digital
//               I/O port.
// Parameters  : DeviceNumber(IN)
//               port(IN)
//				 value(OUT)
// Return      : TRUE (success), FALSE (failure)
//
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG DioGetCurrentDOByte(USHORT DeviceNumber, USHORT port, USHORT * value)
{
    return((ULONG) GetDoStatus(DeviceNumber, port, (USHORT *)value));
}

//------------------------------------------------------------------
// Function    : DioGetCurrentDOBit()
// PURPOSE     : Returns the bit data of digital input from the specified
//               digital I/O port.
// Parameters  : DeviceNumber(IN)
//               port(IN)
//				 bit(IN)
//				 state(OUT)
// Return      : TRUE (success), FALSE (failure)
//
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//------------------------------------------------------------------
ULONG DioGetCurrentDOBit(USHORT DeviceNumber,
    USHORT port, USHORT bit, USHORT * state)
{
    USHORT usMask;
    ULONG  retCode;

    if (bit > 7)
        return((ULONG) InvalidInputParam);

    retCode = GetDoStatus(DeviceNumber, port, (USHORT  *)state);

    if (retCode == 0L)
    {
        usMask = 0x01;
        usMask = usMask << bit;
        *state = *state & usMask;

        if (*state != 0)
            *state = 1;
    }

    return((ULONG) retCode);
}

//------------------------------------------------------------------
// Function    : DoSection()
// PURPOSE     : Writes digital output data to the specified the port.
//               I/O port.
// Parameters  : DeviceNumber(IN)
//               usChlNum(IN)
//				 usmask(IN)
//				 usData(IN)
// Return      : None
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//             DioWritePortByte
//			   DioWriteBit
//------------------------------------------------------------------
ULONG DoSection(USHORT DeviceNumber, USHORT usChlNum,
                USHORT usMask, USHORT usData)
{
    USHORT usBase;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (usChlNum > MaxDioNumber)
        return((ULONG) InvalidPortChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    usData = usData & 0x00FF;
    usMask = usMask & 0x00FF;

    /* set the bits enable by the mask word */
    /* to values specified by the data word */
    /* leaving all other bits in the port unchange */
    usData = (usData & usMask) |
             (usPreState[DeviceNumber][usChlNum] & ~(usMask));

    /* write the desired output state to the hardware */
    outportb(usBase+usChlNum+16, usData);

    /* save current channel output state */
    usPreState[DeviceNumber][usChlNum] = usData;

    return((ULONG) 0L);
}

//------------------------------------------------------------------
// Function    : DiSection()
// PURPOSE     : Returns digital input data from the specified digital
//               I/O port.
// Parameters  : DeviceNumber(IN)
//               usChlNum(IN)
//				 usData(OUT)
// Return      : None
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//             DioReadPortByte
//			   DioReadBit
//------------------------------------------------------------------
ULONG DiSection(USHORT DeviceNumber, USHORT usChlNum, USHORT  * usData)
{
    USHORT usBase;

    // Is DeviceNumber out of range ?
    if (DeviceNumber > MaxBoardNumber)
        return ((ULONG) InvalidDeviceNumber);

    // Is BaseAddr valid ?
    if (usBaseAddr[DeviceNumber] == NULL)
        return ((ULONG) LostBaseAddr);

    // check input param correct ?
    if (usChlNum > MaxDioNumber)
        return((ULONG) InvalidPortChannel);

    // get base address
    usBase = usBaseAddr[DeviceNumber];

    *usData = inportb(usBase+usChlNum+16);

    return((ULONG) 0L);
}

//------------------------------------------------------------------
// Function    : GetDoStatus()
// PURPOSE     : Returns digital output data from the specified digital
//               I/O port or lpDevInfo.
// Parameters  : DeviceNumber(IN)
//               usChlNum(IN)
//				 usData(OUT)
// Return      : None
// Call/Called procedure cross reference :
// Call        Called		     Explanation
//             DioReadPortByte
//			   DioReadBit
//------------------------------------------------------------------
ULONG GetDoStatus(USHORT DeviceNumber, USHORT usChlNum, USHORT * usData)
{
    // check input param correct ?
    if (usChlNum > MaxDioNumber)
        return((ULONG) InvalidPortChannel);

    *usData = usPreState[DeviceNumber][usChlNum];

    return((ULONG) 0L);
}

